<?php
/*
Plugin Name: GF Stripe Extensions
Plugin URI: https://wordpress.org/plugins/gf-stripe-extensions
Version: 2.1.9
Description: Add Stripe functions to Wordpress (Apple Pay) including limit payments and payment recovery to Gravity Forms.
Author: James Low
Author URI: http://jameslow.com
*/

require_once 'cachelib.php'; //MySQL based value cache to cache external apis locally
require_once 'gf-stripe-limit-payments.php'; //Limit stripe payments to certain number
require_once 'gf-stripe-apple-pay.php'; //Apple pay functions
require_once 'gf-stripe-analytics.php'; //Analytics UI and API
require_once 'gf-stripe-payment-recovery.php'; //Payment Recovery
require_once 'gf-stripe-create-entries.php'; //Create entries for missing payments


class GFStripeExtensions {
	public static $PREFIX = 'stripe-extensions';
	public static $GROUP = 'stripe-extensions-settings-group';
	public static $forms = null;
	public static $FIRSTYEAR = 2008; //First year of Gravity Forms
	public static $ValueCache = null;
	public static $roles = null;
	public static $paypaltoken = null;

	private static $_version = null;

	public static function version() {
		if (!self::$_version) {
			if (!function_exists('get_plugin_data')){
				require_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}
			$data = get_plugin_data(__FILE__);
			self::$_version = $data['Version'];
		}
		return self::$_version;
	}

	public static function add_hooks() {
		add_action('plugins_loaded', array('GFStripeExtensions', 'plugins_loaded'));
		add_action('gform_loaded', array('GFStripeExtensions', 'gform_loaded'), 6); //6 so it loads after stripe 5
		add_action('admin_menu', array('GFStripeExtensions', 'admin_menu'));
		add_action('admin_enqueue_scripts', array('GFStripeExtensions', 'include_cssjs'));
		//add_filter('gform_stripe_customer_id', array('GFStripeExtensions', 'gform_stripe_customer_id'), 10, 4);
		//gform_post_payment_completed
		//add_action('gform_post_payment_action', array('GFStripeExtensions', 'post_payment'), 10, 2);
	}

	public static function roles() {
		if (!self::$roles) {
			self::$roles = array(
				'manage_options' => 'Adminstrator',
				'publish_pages' => 'Editor',
				'publish_posts' => 'Author',
				'edit_posts' => 'Contributor',
				'read' => 'Subscriber'
			);
		}
		return self::$roles;
	}
	public static function include_cssjs() {
		if ($_GET['page'] == 'gf-analytics') {
			wp_enqueue_script('jquery');
			wp_enqueue_script('jquery-ui-autocomplete');
		}
	}
	public static function plugins_loaded() {
		self::$ValueCache = new ValueCache(__FILE__, 'gfse');
		self::$ValueCache->enable = self::get_boolean('analytics-cache');
	}
	public static function gform_loaded() {
		require_once 'gf-stripe-extensions-addon.php'; //Stripe addon
		GFAddOn::register('GFStripeExtensionsAddon');
		if (class_exists('GFStripe') && (GFCreateEntries::process_stripe() || self::get_boolean('stripe-allow-feeds'))) {
			GFAddOn::register('GFStripeExt', 'GFStripe');
		}
		if (class_exists('GFPayPal') && GFCreateEntries::process_paypal()) {
			GFAddOn::register('GFPayPalExt', 'GFPayPal');
		}
		if (self::get_boolean('paypal-ignore-hash')) {
			add_filter('gform_paypal_hash_matches', array('GFStripeExtensions', 'gform_paypal_hash_matches'), 10, 4);
		}
	}
	public static function gform_paypal_hash_matches($hash_matches, $entry_id, $hash, $custom_field) {
		return true;
	}
	public static function stripe_event() {
		$body = @file_get_contents('php://input');
		return json_decode($body, true);
	}
	public static function get_role() {
		$role = self::get_option('analytics-role', 'manage_options');
		return $role && $role != '' ? $role : 'manage_options';
	}
	public static function only_role() {
		$role = self::get_option('analytics-only', 'publish_pages');
		return $role && $role != '' ? $role : 'publish_pages';
	}
	public static function admin_menu() {
		//Manage options because only admin should be able to edit settings
		add_options_page('Stripe Extensions', 'Stripe Extensions', 'manage_options', self::$PREFIX, array('GFStripeExtensions', 'options_page'));
		add_action('admin_init', array('GFStripeExtensions', 'admin_init'));
	}
	public static function register_setting($setting, $args = array()) {
		//register_setting(self::$GROUP, self::$PREFIX.'-analytics-key');
		register_setting(self::$GROUP, self::$PREFIX.'-'.$setting, $args);
	}
	public static function admin_init() {
		self::register_setting('analytics-key');
		self::register_setting('analytics-additional');
		self::register_setting('analytics-total');
		self::register_setting('analytics-value');
		self::register_setting('analytics-year', ''.self::$FIRSTYEAR);
		self::register_setting('analytics-cache', '1');
		self::register_setting('analytics-role', 'manage_options');
		self::register_setting('analytics-only', 'publish_pages');
		self::register_setting('stripe-public');
		self::register_setting('stripe-secret');
		self::register_setting('stripe-public-test');
		self::register_setting('stripe-secret-test');
		self::register_setting('stripe-form');
		self::register_setting('stripe-test-webhooks');
		self::register_setting('stripe-allow-feeds');
		self::register_setting('stripe-create-customer');
		self::register_setting('getresponse-apikey');
		self::register_setting('getresponse-domain');
		self::register_setting('getresponse-enterprise');
		self::register_setting('virtuous-token');
		self::register_setting('virtuous-campaigns');
		self::register_setting('applepay-form');
		self::register_setting('paypal-form');
		self::register_setting('paypal-ignore-hash');
		self::register_setting('paypal-clientid');
		self::register_setting('paypal-secret');
		self::register_setting('paypal-clientid-test');
		self::register_setting('paypal-secret-test');
		self::register_setting('payment-type');
		self::register_setting('recovery-enable');
		self::register_setting('recovery-url');
		self::register_setting('recovery-address');
		self::register_setting('recovery-name');
		self::register_setting('recovery-subject');
		self::register_setting('recovery-bcc');
		self::register_setting('recovery-template', 'Dear [fullname],

		Thank you for your continued support for [sitename].
		
		Your recent recurring payment has [error]. Please update your billing information at the link below:
		<a href="[link]">[link]</a>
		
		Thank You,
		[sitename]');
	}
	public static function get_forms() {
		if (!self::$forms) {
			self::$forms = RGFormsModel::get_forms(1, 'title');
		}
		return self::$forms;
	}
	public static function feed_select($option, $title, $none = '(None)') {
		if (class_exists('RGFormsModel')) {
			$forms = self::get_forms();
			$values = array();
			foreach ($forms as $form) {
				$values[$form->id] = $form->title;
			}
		?>
			<tr valign="top"><th scope="row"><?php echo $title; ?></th>
			<td colspan="2"><?php echo self::select($option, $values, $none); ?></td></tr>
<?php 	}
	}
	public static function select($option, $values, $none = null) {
		$current = self::get_option($option);
		?>
		<select name="<?php echo self::$PREFIX.'-'.$option; ?>">
			<?php
			echo $none ? '<option value="0">'.$none.'</option>' : '';
			foreach ($values as $key => $name) {
				echo '<option value="'.$key.'"'.($current==$key?' selected="selected"':'').'>'.esc_html($name).'</option>';
			}
			?>
		</select>
		<?php
	}
	public static function years() {
		$years = array();
		$year = intval(date("Y"));
		$firstyear = self::$FIRSTYEAR;
		$dates = array();
		for ($i=$year; $i>=$firstyear; $i--) {
			$dates[''.$i] = ''.$i;
		}
		return $dates;
	}
	public static function checkbox($id, $label) {
		return '<input id="'.$id.'" type="checkbox" name="'.self::$PREFIX.'-'.$id.'" value="1" '.checked(1 == self::get_boolean($id), true, false).' />
		<label for="'.$id.'">'.$label.'</label>';
	}
	public static function textarea($id) {
		return '<textarea id="'.$id.'" name="'.self::$PREFIX.'-'.$id.'" cols="45" style="height:120px;">'.esc_html(self::get_option($id)).'</textarea>';
	}
	public static function options_page() { ?>
		<div class="wrap">
		<h1>Stripe Extensions</h1>
		<form method="post" action="options.php">
		<?php
		settings_fields(self::$GROUP);
		do_settings_sections(self::$GROUP);
		$fromgravity = class_exists('GFStripe');
		?>
		
		<table class="form-table">
			<tr valign="top">
				<th colspan="2"><h2>Analytics</h2></th>
			</tr>
			<tr valign="top">
				<th scope="row">API Key</th>
				<td><?php echo self::setting('analytics-key','API Key'); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Additional Keys</th>
				<td><?php echo self::textarea('analytics-additional'); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Analytics Grouping</th>
				<td><?php echo self::setting('analytics-total','(Payments)'); ?>
				<?php echo self::setting('analytics-value','(Value)'); ?>
				</td>
			</tr>
			<tr valign="top">
				<th scope="row">First Year</th>
				<td><?php self::select('analytics-year', self::years()); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Caching</th>
				<td><?php echo self::checkbox('analytics-cache', 'Enable Analytics Cache'); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Full Access Role</th>
				<td><?php self::select('analytics-role', self::roles()); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Analytics Role</th>
				<td><?php self::select('analytics-only', self::roles()); ?></td>
			</tr>
			<tr valign="top">
				<th colspan="2"><h2>Gravity Forms</h2></th>
			</tr>
			<tr valign="top">
				<th scope="row">Payment Type</th>
				<td><?php echo self::checkbox('payment-type', 'Backfill for recurring payments'); ?></td>
			</tr>
			<?php
			self::feed_select('applepay-form', 'ApplePay Form', "(Don't log ApplePay)");
			self::feed_select('stripe-form', 'Stripe Unknown Form', "(Don't create entries)");
			self::feed_select('paypal-form', 'PayPal Unknown Form', "(Don't create entries)");
			?>
			<tr valign="top">
				<th colspan="2"><h2>PayPal</h2></th>
			</tr>
			<tr valign="top">
				<th scope="row">PayPal Live</th>
				<td><?php echo self::setting('paypal-clientid','Client Id'); ?>
				<?php echo self::setting('paypal-secret','Secret',true); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">PayPal Sandbox</th>
				<td><?php echo self::setting('paypal-clientid-test','Sandbox Client Id'); ?>
				<?php echo self::setting('paypal-secret-test','Sandbox Secret',true); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">PayPal Entry Hash</th>
				<td><?php echo self::checkbox('paypal-ignore-hash', 'Ignore Hash'); ?></td>
			</tr>
			<tr valign="top">
				<th colspan="2"><h2>Stripe</h2></th>
			</tr>
			<tr valign="top">
				<th scope="row">Stripe Live</th>
				<td><?php echo self::setting('stripe-public',$fromgravity?'(Public From Gravity Forms)':'Public Key'); ?>
				<?php echo self::setting('stripe-secret',$fromgravity?'(Secret From Gravity Forms)':'Secret Key',true); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Stripe Test</th>
				<td><?php echo self::setting('stripe-public-test',$fromgravity?'(Public From Gravity Forms)':'Public Key'); ?>
				<?php echo self::setting('stripe-secret-test',$fromgravity?'(Secret From Gravity Forms)':'Secret Key',true); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Enable Test Webhooks</th>
				<td><?php echo self::checkbox('stripe-test-webhooks', 'Enable stripe test webhooks (debugging)'); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Allow Legacy Feeds</th>
				<td><?php echo self::checkbox('stripe-allow-feeds', 'Allow Legacy Card Feeds'); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Create Customer</th>
				<td><?php echo self::checkbox('stripe-create-customer', 'Create for Products & Services'); ?></td>
			</tr>

			<tr valign="top">
				<th colspan="2"><h2>Get Response</h2></th>
			</tr>
			<tr valign="top">
				<th scope="row">Api Settings</th>
				<td><?php echo self::setting('getresponse-apikey','Api Key'); ?>
				<?php echo self::setting('getresponse-domain','Domain'); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Enterprise</th>
				<td><?php echo self::checkbox('getresponse-enterprise', 'Enterprise Account'); ?></td>
			</tr>

			<tr valign="top">
				<th colspan="2"><h2>Virtuous</h2></th>
			</tr>
			<tr valign="top">
				<th scope="row">Refresh Token</th>
				<td><?php echo self::setting('virtuous-token','Refresh Token'); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Default Campaigns</th>
				<td><?php echo self::textarea('virtuous-campaigns'); ?></td>
			</tr>

			<tr valign="top">
				<th colspan="2"><h2>Payment Recovery</h2></th>
			</tr>
			<tr valign="top">
				<th scope="row">Recovery Url</th>
				<td>
				<?php echo self::setting('recovery-url','(URL)'); ?>
				<?php echo self::checkbox('recovery-enable', 'Enable Recovery'); ?></td>
			</tr>
			<tr valign="top">
				<th scope="row">Recovery Email</th>
				<td>
					<?php echo self::setting('recovery-subject','(Subject)'); ?>
					<?php echo self::setting('recovery-bcc','(BCC Address)'); ?>
				</td>
			</tr>
			<tr valign="top">
				<th scope="row">Email Address</th>
				<td>
					<?php echo self::setting('recovery-name','(From Name)'); ?>
					<?php echo self::setting('recovery-address','(From Address)'); ?>
				</td>
			</tr>
			<tr valign="top">
				<th scope="row">Email Template
				<div style="font-weight:normal;"><br />Tags:<br />
					[fullname]<br />
					[link]<br />
					[error]<br />
					[email]<br />
					[sitename]<br />
				</div>
				</th>
				<td>
					<?php 
						wp_editor(self::get_option('recovery-template'), 'stripe-extensions-recovery-template', $settings = array('textarea_rows'=> '10') );
					?>
				</td>
			</tr>
			
		</table>
		<?php submit_button(); ?>
		</form><?php
	}
	public static function setting($id, $placeholder = '', $password = false) {
		return '<input placeholder="'.$placeholder.'" type="'.($password?'password':'text').'" name="'.self::$PREFIX.'-'.$id.'" value="'.esc_attr(self::get_option($id)).'" />';
	}
	public static function get_boolean($id) {
		return 1 == self::get_option($id);
	}
	public static function get_option($id, $default = false) {
		$value = get_option(self::$PREFIX.'-'.$id, $default);
		if ($id == 'analytics-key') {
			if (!$value) {
				$value = md5(wp_salt() . time());
				self::set_option($id, $value);
			}
		} elseif ($id == 'analytics-year') {
			if (!$value || $value == '') {
				$value = 2008;
				self::set_option($id, $value);
			}
		}
		return $value;
	}
	public static function get_campaigns() {
		$values = trim(self::get_option('virtuous-campaigns', ''));
		$array = explode("\n", $values);
		$results = array();
		foreach ($array as $row) {
			if (trim($row) != '') {
				$results[] = $row;
			}
		}
		return $results;
	}
	public static function get_stripe_key($id, $stripe) {
		$value = self::get_stripe_setting($stripe);
		return $value !== null ? $value : self::get_option($id);
	}
	public static function get_stripe_setting($stripe) {
		if (class_exists('GFStripe')) {
			return GFStripe::get_instance()->get_plugin_setting($stripe);
		}
	}
	public static function stripe_test() {
		return self::get_stripe_setting('api_mode') != 'live';
	}
	public static function get_stripe_live_public() {
		return self::get_stripe_key('stripe-public', 'live_publishable_key');
	}
	public static function get_stripe_live_secret() {
		return self::get_stripe_key('stripe-secret', 'live_secret_key');
	}
	public static function get_stripe_test_public() {
		return self::get_stripe_key('stripe-public-test', 'test_publishable_key');
	}
	public static function get_stripe_test_secret() {
		return self::get_stripe_key('stripe-secret-test', 'test_secret_key');
	}
	public static function set_option($id, $value) {
		update_option(self::$PREFIX.'-'.$id, $value);
	}
	public static function post_payment($entry, $action) {
		//This code currently isn't being used, as we decided to just map the card field, rather than populate during submission
		if (class_exists('GFStripe') && isset($entry['payment_method']) && $entry['payment_method'] != '') {
			$stripe = GFStripe::get_instance();
			if (isset($_POST['stripe_response'])) {
				$response = $stipe->get_stripe_js_response();
				$stripe->log_debug( __METHOD__ . print_r($entry, true));
				$stripe->log_debug( __METHOD__ . print_r($action, true));
				$stripe->log_debug( __METHOD__ . print_r($response, true));
				if (isset($response->token) && isset($response->token->card)) {
					$entry['payment_method'] = $response->token->card->brand;
					GFAPI::update_entry($entry);
				}
			}
		}
	}
	public static function stripe($path, $params = null, $cache = null) {
		if ($cache) {
			$result = self::$ValueCache->get($cache);
			if ($result) {
				return json_decode($result, true);
			}
		}

		require_once 'httplib.php';
		$secret = self::stripe_test() ? self::get_stripe_test_secret() : self::get_stripe_live_secret();
		if ($secret == '' || $secret == null) {
			throw new Exception('Stripe secret key has not been set.');
		}
		$http = new HTTPRequest('https://api.stripe.com/v1'.$path);
		$headerarray = array('Authorization' => 'Bearer ' . $secret);
		$result = $params ? $http->Post($params, true, $headerarray) : $http->Get(array(), true, $headerarray);
		if ($result['http']['code'] == 200) {
			$body = $result['body'];
			if ($cache) {
				self::$ValueCache->put($cache, $body);
			}
			return json_decode($body, true);
		} else {
			$error = json_decode($result['body'], true);
			error_log(print_r($error, true));
			throw new Exception('Stripe Error: '.$error['error']['message']);
		}
	}
	public static function virtous_token() {
		require_once 'httplib.php';
		$http = new HTTPRequest('https://api.virtuoussoftware.com/Token');
		$token = $http->Post(array('grant_type' => 'refresh_token', 'refresh_token' => self::get_option('virtuous-token')), true);
		$token = json_decode($token['body']);
		return $token->access_token;
	}
	public static function virtuous($uri, $paramsbody = array(), $method = 'GET', $cache = null) {
		if ($cache) {
			$result = self::$ValueCache->get($cache);
			if ($result) {
				return json_decode($result);
			}
		}
		require_once 'httplib.php';
		$headerarray = array(
			'Authorization' => 'Bearer '.self::virtous_token()
		);
		$http = new HTTPRequest('https://api.virtuoussoftware.com/api'.$uri);
		if ($method == 'POST') {
			$headerarray['Content-Type'] = 'application/json';
			$result = $http->Post($paramsbody, true, $headerarray);
		} else {
			$result = $http->Query($method, $paramsbody, true, $headerarray);
		}
		$body = $result['body'];
		if ($cache) {
			self::$ValueCache->put($cache, $body);
		}
		return json_decode($body);
	}
	public static function query($sql, $cache = null) {
		global $wpdb;
		if ($cache) {
			$result = self::$ValueCache->get($cache);
			if ($result) {
				return json_decode($result);
			}
		}
		$results = $wpdb->get_results($sql);
		if ($cache) {
			self::$ValueCache->put($cache, json_encode($results));
		}
		return $results;
	}
	public static function paypal_token() {
		if (self::$paypaltoken == null) {
			require_once 'httplib.php';
			$apiurl = 'https://api.paypal.com';
			$http = new HTTPRequest($apiurl.'/v1/oauth2/token');
			//echo self::get_option('paypal-clientid') . '<br/>';
			//echo self::get_option('paypal-secret') . '<br/>';
			$headerarray = array(
				'Accept' => 'application/json',
				'Accept-Language' => 'en_US',
				'Authorization' => 'Basic '.base64_encode(self::get_option('paypal-clientid').':'.self::get_option('paypal-secret'))
			);
			$params = array(
				'grant_type' => 'client_credentials'
			);
			$result = $http->Post($params, true, $headerarray);
			$json = json_decode($result['body']);
			return $json->access_token;
		}
		return self::$paypaltoken;
	}
	public static function paypal($uri, $paramsbody = array(), $method = 'GET', $cache = null) {
		if ($cache) {
			$result = self::$ValueCache->get($cache);
			if ($result) {
				return json_decode($result);
			}
		}
		require_once 'httplib.php';
		$apiurl = 'https://api.paypal.com';
		$refresh = self::get_option('virtuous-token');
		$http = new HTTPRequest($apiurl.$uri);
		$headerarray = array(
			'Content-Type' => 'application/json',
			'Authorization' => 'Bearer '.self::paypal_token()
		);
		$http = new HTTPRequest($apiurl.$uri);
		$result = $http->Query($method, $paramsbody, true, $headerarray);
		$body = $result['body'];
		if ($cache) {
			self::$ValueCache->put($cache, $body);
		}
		return json_decode($body);
	}
	public static function gform_stripe_customer_id($customer_id, $feed, $entry, $form) {
		if (self::get_boolean('stripe-create-customer') && !$customer_id && class_exists('\Stripe\Customer')) {
			$details = GFAnalytics::map_entry($entry, $form);
			//We could still create a customer if there is no email, as it would feed through data to Virtuous, but might create duplicates if can't check if it exists by email first
			if ($details && $details->email) {
				$customer = \Stripe\Customer::all(array('email' => $details->email));
				if ($customer && $customer->data && count($customer->data) > 0) {
					return $customer->data[0]->id;
				} else {
					$meta = array(
						'email' => $details->email,
						'name' => trim($details->firstname.' '.$details->lastname),
						'phone' => $details->phone,
						'address' =>  array(
							'line1' => $details->address1,
							'line2' => $details->address2,
							'city' => $details->city,
							'state' => $details->state,
							'postal_code' => $details->zip,
							//'country' => $details->country //TODO: convert to country code
						)
					);
					$customer = \Stripe\Customer::create($meta);
					if ($customer && !is_wp_error($customer)) {
						return $customer->id;
					}
				}
			}
		}
		return $customer_id;
	}
}
GFStripeExtensions::add_hooks();